]> git.cameronkatri.com Git - mandoc.git/blobdiff - mdoc.c
Commit and commit message by deraadt@:
[mandoc.git] / mdoc.c
diff --git a/mdoc.c b/mdoc.c
index a1e3a0f9955a4435b13bd7d92a70e578f23de6fc..475a6aa0ddb3cac71309fe5fe5450f1ce110c825 100644 (file)
--- a/mdoc.c
+++ b/mdoc.c
@@ -1,7 +1,7 @@
-/*     $Id: mdoc.c,v 1.263 2017/04/29 12:45:41 schwarze Exp $ */
+/* $Id: mdoc.c,v 1.275 2020/04/06 10:16:17 schwarze Exp $ */
 /*
+ * Copyright (c) 2010, 2012-2018, 2020 Ingo Schwarze <schwarze@openbsd.org>
  * Copyright (c) 2008, 2009, 2010, 2011 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2010, 2012-2017 Ingo Schwarze <schwarze@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,6 +14,8 @@
  * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
  * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ *
+ * Top level and utility functions of the mdoc(7) parser for mandoc(1).
  */
 #include "config.h"
 
@@ -79,13 +81,6 @@ mdoc_parseln(struct roff_man *mdoc, int ln, char *buf, int offs)
            mdoc_ptext(mdoc, ln, buf, offs);
 }
 
-void
-mdoc_macro(MACRO_PROT_ARGS)
-{
-       assert(tok >= MDOC_Dd && tok < MDOC_MAX);
-       (*mdoc_macros[tok].fp)(mdoc, tok, line, ppos, pos, buf);
-}
-
 void
 mdoc_tail_alloc(struct roff_man *mdoc, int line, int pos, enum roff_tok tok)
 {
@@ -162,15 +157,6 @@ mdoc_elem_alloc(struct roff_man *mdoc, int line, int pos,
        mdoc->next = ROFF_NEXT_CHILD;
 }
 
-void
-mdoc_node_relink(struct roff_man *mdoc, struct roff_node *p)
-{
-
-       roff_node_unlink(mdoc, p);
-       p->prev = p->next = NULL;
-       roff_node_append(mdoc, p);
-}
-
 /*
  * Parse free-form text, that is, a line that does not begin with the
  * control character.
@@ -179,6 +165,7 @@ static int
 mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
 {
        struct roff_node *n;
+       const char       *cp, *sp;
        char             *c, *ws, *end;
 
        n = mdoc->last;
@@ -195,7 +182,8 @@ mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
            (n->parent != NULL && n->parent->tok == MDOC_Bl &&
             n->parent->norm->Bl.type == LIST_column)) {
                mdoc->flags |= MDOC_FREECOL;
-               mdoc_macro(mdoc, MDOC_It, line, offs, &offs, buf);
+               (*mdoc_macro(MDOC_It)->fp)(mdoc, MDOC_It,
+                   line, offs, &offs, buf);
                return 1;
        }
 
@@ -224,7 +212,7 @@ mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
                         * Strip trailing tabs in literal context only;
                         * outside, they affect the next line.
                         */
-                       if (MDOC_LITERAL & mdoc->flags)
+                       if (mdoc->flags & ROFF_NOFILL)
                                continue;
                        break;
                case '\\':
@@ -241,19 +229,32 @@ mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
        *end = '\0';
 
        if (ws)
-               mandoc_msg(MANDOCERR_SPACE_EOL, mdoc->parse,
-                   line, (int)(ws-buf), NULL);
-
-       if (buf[offs] == '\0' && ! (mdoc->flags & MDOC_LITERAL)) {
-               mandoc_msg(MANDOCERR_FI_BLANK, mdoc->parse,
-                   line, (int)(c - buf), NULL);
-
-               /*
-                * Insert a `sp' in the case of a blank line.  Technically,
-                * blank lines aren't allowed, but enough manuals assume this
-                * behaviour that we want to work around it.
-                */
-               roff_elem_alloc(mdoc, line, offs, MDOC_sp);
+               mandoc_msg(MANDOCERR_SPACE_EOL, line, (int)(ws - buf), NULL);
+
+       /*
+        * Blank lines are allowed in no-fill mode
+        * and cancel preceding \c,
+        * but add a single vertical space elsewhere.
+        */
+
+       if (buf[offs] == '\0' && (mdoc->flags & ROFF_NOFILL) == 0) {
+               switch (mdoc->last->type) {
+               case ROFFT_TEXT:
+                       sp = mdoc->last->string;
+                       cp = end = strchr(sp, '\0') - 2;
+                       if (cp < sp || cp[0] != '\\' || cp[1] != 'c')
+                               break;
+                       while (cp > sp && cp[-1] == '\\')
+                               cp--;
+                       if ((end - cp) % 2)
+                               break;
+                       *end = '\0';
+                       return 1;
+               default:
+                       break;
+               }
+               mandoc_msg(MANDOCERR_FI_BLANK, line, (int)(c - buf), NULL);
+               roff_elem_alloc(mdoc, line, offs, ROFF_sp);
                mdoc->last->flags |= NODE_VALID | NODE_ENDED;
                mdoc->next = ROFF_NEXT_SIBLING;
                return 1;
@@ -261,7 +262,7 @@ mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
 
        roff_word_alloc(mdoc, line, offs, buf+offs);
 
-       if (mdoc->flags & MDOC_LITERAL)
+       if (mdoc->flags & ROFF_NOFILL)
                return 1;
 
        /*
@@ -278,16 +279,21 @@ mdoc_ptext(struct roff_man *mdoc, int line, char *buf, int offs)
        for (c = buf + offs; c != NULL; c = strchr(c + 1, '.')) {
                if (c - buf < offs + 2)
                        continue;
-               if (end - c < 4)
+               if (end - c < 3)
                        break;
-               if (isalpha((unsigned char)c[-2]) &&
-                   isalpha((unsigned char)c[-1]) &&
-                   c[1] == ' ' &&
-                   isupper((unsigned char)(c[2] == ' ' ? c[3] : c[2])) &&
-                   (c[-2] != 'n' || c[-1] != 'c') &&
-                   (c[-2] != 'v' || c[-1] != 's'))
-                       mandoc_msg(MANDOCERR_EOS, mdoc->parse,
-                           line, (int)(c - buf), NULL);
+               if (c[1] != ' ' ||
+                   isalnum((unsigned char)c[-2]) == 0 ||
+                   isalnum((unsigned char)c[-1]) == 0 ||
+                   (c[-2] == 'n' && c[-1] == 'c') ||
+                   (c[-2] == 'v' && c[-1] == 's'))
+                       continue;
+               c += 2;
+               if (*c == ' ')
+                       c++;
+               if (*c == ' ')
+                       c++;
+               if (isupper((unsigned char)(*c)))
+                       mandoc_msg(MANDOCERR_EOS, line, (int)(c - buf), NULL);
        }
 
        return 1;
@@ -315,8 +321,7 @@ mdoc_pmacro(struct roff_man *mdoc, int ln, char *buf, int offs)
        if (sz == 2 || sz == 3)
                tok = roffhash_find(mdoc->mdocmac, buf + sv, sz);
        if (tok == TOKEN_NONE) {
-               mandoc_msg(MANDOCERR_MACRO, mdoc->parse,
-                   ln, sv, buf + sv - 1);
+               mandoc_msg(MANDOCERR_MACRO, ln, sv, "%s", buf + sv - 1);
                return 1;
        }
 
@@ -346,17 +351,17 @@ mdoc_pmacro(struct roff_man *mdoc, int ln, char *buf, int offs)
         */
 
        if ('\0' == buf[offs] && ' ' == buf[offs - 1])
-               mandoc_msg(MANDOCERR_SPACE_EOL, mdoc->parse,
-                   ln, offs - 1, NULL);
+               mandoc_msg(MANDOCERR_SPACE_EOL, ln, offs - 1, NULL);
 
        /*
* If an initial macro or a list invocation, divert directly
* into macro processing.
         */
 
        n = mdoc->last;
-       if (n == NULL || tok == MDOC_It || tok == MDOC_El) {
-               mdoc_macro(mdoc, tok, ln, sv, &offs, buf);
+       if (n == NULL || tok == MDOC_It || tok == MDOC_El ||
+           roff_tok_transparent(tok)) {
+               (*mdoc_macro(tok)->fp)(mdoc, tok, ln, sv, &offs, buf);
                return 1;
        }
 
@@ -372,13 +377,13 @@ mdoc_pmacro(struct roff_man *mdoc, int ln, char *buf, int offs)
            (n->parent != NULL && n->parent->tok == MDOC_Bl &&
             n->parent->norm->Bl.type == LIST_column)) {
                mdoc->flags |= MDOC_FREECOL;
-               mdoc_macro(mdoc, MDOC_It, ln, sv, &sv, buf);
+               (*mdoc_macro(MDOC_It)->fp)(mdoc, MDOC_It, ln, sv, &sv, buf);
                return 1;
        }
 
        /* Normal processing of a macro. */
 
-       mdoc_macro(mdoc, tok, ln, sv, &offs, buf);
+       (*mdoc_macro(tok)->fp)(mdoc, tok, ln, sv, &offs, buf);
 
        /* In quick mode (for mandocdb), abort after the NAME section. */
 
@@ -426,12 +431,3 @@ mdoc_isdelim(const char *p)
 
        return DELIM_NONE;
 }
-
-void
-mdoc_validate(struct roff_man *mdoc)
-{
-
-       mdoc->last = mdoc->first;
-       mdoc_node_validate(mdoc);
-       mdoc_state_reset(mdoc);
-}
@@ -346,17 +351,17 @@ mdoc_pmacro(struct roff_man *mdoc, int ln, char *buf, int offs)
 
        /*
-        * If an initial macro or a list invocation, divert directly
-        * into macro processing.
+        * If an initial or transparent macro or a list invocation,
+        * divert directly into macro processing.
         */
 